from requests import request

from .schemas import (FeedbackSchema, FeedbackResponse, TrainResponse,
                      RankResponse, RecommendResponse, RecommendSchema,
                      InferenceEncoderSchema, InferenceEncoderResponse,
                      InferenceCrossSchema, InferenceCrossResponse, RankSchema)


class Client:
    def __init__(self, base_url: str):
        self.base_url = base_url

    def feedback(self, feedback_data: FeedbackSchema) -> FeedbackResponse:
        """
        See: https://docs.metarank.ai/reference/api#feedback
        
        :param feedback_data: 
        :return: 
        """
        response = self.request("POST", "feedback", feedback_data.dict(exclude_none=True))
        return FeedbackResponse(**response)
    
    def health_check(self) -> bool:
        """
        :return: Whether the metarank service is ready or not 
        """

        response = self.raw_request('GET', 'health')
        return response.ok
    
    def inference_encoder(self, name: str, inference_encoder_data: InferenceEncoderSchema) -> InferenceEncoderResponse:
        """
        See: https://docs.metarank.ai/reference/api#inference-with-llms

        :param name: 
        :param inference_encoder_data: 
        :return: 
        """

        response = self.request("POST", f"inference/encoder/{name}", inference_encoder_data.dict(exclude_none=True))
        return InferenceEncoderResponse(**response)
    
    def inference_cross(self, name: str, inference_cross_data: InferenceCrossSchema) -> InferenceCrossResponse:
        """
        See: https://docs.metarank.ai/reference/api#inference-with-llms

        :param name: 
        :param inference_cross_data: 
        :return: 
        """

        response = self.request("POST", f"inference/cross/{name}", inference_cross_data.dict(exclude_none=True))
        return InferenceCrossResponse(**response)
    
    def metrics(self) -> str:
        """
        See: https://docs.metarank.ai/reference/api#prometheus-metrics

        :return: 
        """
        
        response = self.raw_request("GET", "metrics")
        return response.text
    
    def rank(self, model_name: str, rank_data: RankSchema, explain: bool = False) -> RankResponse:
        """
        See: https://docs.metarank.ai/reference/api#ranking

        :param model_name: 
        :param explain: 
        :return: 
        """

        endpoint = f"rank/{model_name}"
        if explain:
            endpoint = f"{endpoint}?explain=true"

        response = self.request("POST", endpoint, data=rank_data.dict(exclude_none=True))
        return RankResponse(**response)
    
    def recommend(self, model_name: str, data: RecommendSchema) -> RecommendResponse:
        """
        See: https://docs.metarank.ai/reference/api#recommendations

        :param model_name: 
        :param data: 
        :return: 
        """

        endpoint = f"recommend/{model_name}"
        
        response = self.request("POST", endpoint, data=data)
        return RecommendResponse(**response)
    
    def train(self, model_name: str) -> TrainResponse:
        """
        See: https://docs.metarank.ai/reference/api#train

        :param model_name: 
        :return: 
        """

        response = self.request("POST", f"train/{model_name}")
        return TrainResponse(**response)

    def request(self, method, endpoint, data=None):
        headers = {'Content-Type': 'application/json'}
        response = self.raw_request(method, endpoint, json=data, headers=headers)
        self._handle_error(response)
        return response.json()

    def raw_request(self, method: str, endpoint: str, **kwargs):
        url = f"{self.base_url}/{endpoint}"
        return request(method, url, **kwargs)

    @staticmethod
    def _handle_error(response):
        if response.status_code != 200:
            raise Exception(f"Error {response.status_code}: {response.text}")



from typing import Literal, Union, Optional

from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


class IdSchema(BaseModel):
    id: str


class BaseSchema(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )


class FieldSchema(BaseSchema):
    name: str
    value: Union[bool, str, int, float, list[str], list[int], list[float]]


class ItemSchema(BaseSchema):
    id: str
    fields: Optional[list[FieldSchema]] = None
    label: Optional[str] = None


class FeedbackSchema(BaseSchema):
    event: Literal["item", "user", "interaction", "ranking"]
    id: str
    item: Optional[str] = None
    items: Optional[list[ItemSchema]] = None
    fields: Optional[list[FieldSchema]] = None
    user: Optional[str] = None
    session: Optional[str] = None
    timestamp: Union[int, str] 


class FeedbackResponse(BaseSchema):
    accepted: int
    status: str
    tookMillis: int
    updated: int


class FeatureSchema(BaseSchema):
    name: str
    weight: list[float]
    

class IterationSchema(BaseSchema):
    id: str
    millis: int
    test_metric: float
    train_metric: float


class RankSchema(BaseSchema):
    id: str
    user: Optional[str] = None
    session: Optional[str] = None
    timestamp: Union[int, str]
    fields: Optional[list[FieldSchema]] = None
    items: list[ItemSchema]

    
class ItemResponse(BaseSchema):
    item: str
    score: float
    features: Optional[list[FieldSchema]] = None
    

class RankResponse(BaseSchema):
    took: int
    items: list[ItemResponse]


class RecommendSchema(BaseSchema):
    count: int
    user: str
    items: list[str]
    
    
class RecommendResponse(BaseSchema):
    took: int
    items: list[ItemResponse]


class TrainResponse(BaseSchema):
    features: list[FeatureSchema]
    iterations: list[IterationSchema]
    size_bytes: int


class InferenceEncoderSchema(BaseSchema):
    texts: list[str]


class InferenceEncoderResponse(BaseSchema):
    took: int
    embeddings: list[list[float]]


class QueryDocumentPairSchema(BaseSchema):
    query: str
    text: str


class InferenceCrossSchema(BaseSchema):
    input: list[QueryDocumentPairSchema]


class InferenceCrossResponse(BaseSchema):
    took: int
    scores: list[float]
